"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createCoreMutationsProvider = void 0;
const chalk_1 = __importDefault(require("chalk"));
const LazyCache_1 = require("../../services/LazyCache");
const FileInfoCache_1 = require("../../shared/FileInfoCache");
const maps_1 = require("../../shared/maps");
const NameGenerator_1 = require("../../shared/NameGenerator");
const collectFilteredNodes_1 = require("../collectFilteredNodes");
const createFileNamesAndServices_1 = require("../createFileNamesAndServices");
const findMutationsInFile_1 = require("../findMutationsInFile");
/**
 * Creates a mutations provider that runs the core mutations within TypeStat.
 *
 * @param options   Parsed runtime options for TypeStat.
 * @param allModifiedFileNames   Set to mark names of all files that were modified.
 */
const createCoreMutationsProvider = (options, allModifiedFiles) => {
    const fileNamesAndServicesCache = createFileNamesAndServicesCache(options);
    let lastFileIndex = -1;
    return () => __awaiter(void 0, void 0, void 0, function* () {
        const startTime = Date.now();
        const fileMutations = new Map();
        const { fileNames, services } = fileNamesAndServicesCache.get();
        const waveStartedFromBeginning = lastFileIndex <= 0;
        let addedMutations = 0;
        for (lastFileIndex = lastFileIndex + 1; lastFileIndex < fileNames.length; lastFileIndex += 1) {
            const fileName = fileNames[lastFileIndex];
            const sourceFile = services.program.getSourceFile(fileName);
            if (sourceFile === undefined) {
                options.output.stderr(`Could not find TypeScript source file for '${fileName}'.`);
                continue;
            }
            const filteredNodes = collectFilteredNodes_1.collectFilteredNodes(options, sourceFile);
            const foundMutations = yield findMutationsInFile_1.findMutationsInFile({
                fileInfoCache: new FileInfoCache_1.FileInfoCache(filteredNodes, services, sourceFile),
                filteredNodes,
                nameGenerator: new NameGenerator_1.NameGenerator(sourceFile.fileName),
                options,
                services,
                sourceFile,
            });
            if (foundMutations !== undefined && foundMutations.length !== 0) {
                addedMutations += foundMutations.length;
                fileMutations.set(fileName, foundMutations);
            }
            if (addedMutations > 100 || (addedMutations !== 0 && Date.now() - startTime > 10000)) {
                break;
            }
        }
        if (lastFileIndex === fileNames.length) {
            lastFileIndex = -1;
            // Only recreate the language service once we've visited every file
            // This way we don't constantly re-scan many of the source files each wave
            // Eventually it would be nice to support incremental updates
            // See https://github.com/JoshuaKGoldberg/TypeStat/issues/36
            fileNamesAndServicesCache.clear();
        }
        for (const fileName of fileMutations.keys()) {
            allModifiedFiles.add(fileName);
        }
        return {
            fileMutations: waveStartedFromBeginning && fileMutations.size === 0 ? undefined : maps_1.convertMapToObject(fileMutations),
        };
    });
};
exports.createCoreMutationsProvider = createCoreMutationsProvider;
const createFileNamesAndServicesCache = (options) => {
    return new LazyCache_1.LazyCache(() => {
        options.output.stdout(chalk_1.default.grey("Preparing language services to visit files..."));
        const { fileNames, services } = createFileNamesAndServices_1.createFileNamesAndServices(options);
        options.output.stdout(chalk_1.default.grey(`Prepared language services for ${fileNames.length} files...`));
        return { fileNames, services };
    });
};
//# sourceMappingURL=createCoreMutationsProvider.js.map